open Interval;;
open Box

type variable = int 
  (* int only known to prune.mli, 
     rest of the world only knows: 
     * variables in constraint are ordered
     * when pruning with box: box coordinates correspond to variables according to their order 
  *)

type predicateTernary = Addition | Multiplication | Subtraction;;
type predicateBinary = Sin | Cos | Tan | ArcSin | ArcCos | ArcTan | Exp | Log | Square | UMinus;;
type predicateUnary = Equal | LessEqual | Less | NotEqual | Greater | GreaterEqual ;;

type primTernary = { predicate3: predicateTernary; var31: variable; var32: variable; var33: variable };;
                   (* predicate3(var31, var32)= var33 *)
type primBinary = { predicate2: predicateBinary; var21: variable; var22: variable; };;
                   (* predicate2(var21)= var22 *) 
type primUnary = { predicate1: predicateUnary; var1: variable };;
type primPower = {var01:variable;n:int;var02:variable};;
type isElement = { var: variable; interval: interval;};;
type posSemDef = variable array array 

type constr  =   
    BoolVar of int     (* a variable ranging over Booleans *)
  | UserDefined of (box -> unit) (* pruning *) * (int -> bool) (* contains variable, is allowed to overapproximate! *)
  | BoolIntv of Bool_intv.bintv  
  | Ternary of primTernary  
  | Binary of primBinary  
  | Unary of primUnary  
  | Power of primPower  
  | IsElement of isElement  
  | SubSet of subset  (* meta-constraint *)
  | PosSemDef of posSemDef  
  | Atomic of atomic   (* existentially quantified conjunction resulting from atomic constraint, optimization: merge with conjunction *)    
  | Conjunction of connective  
  | Disjunction of connective 
  | UnivQuant of quantification 
  | ExistQuant of quantification
  | Redundant of redundant (* redundant constraints that imply a given one *)
  | UnivQuantBound of quantbound 
      (* a block of universally quantified variables that implicitely are 
         bounded by the bound of the corresponding free variables;
         for disproving, the midpoint is taken as a sample
         (used for the implementation of mean-value constraints);
	 the quantified constraint is assumed to be such that neg (forall phi) is 
         equivalent to forall (neg phi). Hence the we can take the sample both in inner
         and outer pruning.
 *)
and quantification = 
    { qbound: box; 
      subconstr: constr; 
      mutable outer_changed: int list; (* variables for which qbound changed since last outer pruning, or never outer pruned, global index (not in qbound) *)
      mutable inner_changed: int list; (* variables for which qbound changed since last inner pruning, or never outer pruned, global index (not in qbound) *)
              (* above lists are just assigned, but not changed, so they can be shared, use bitvector instead? *)
      mutable qnext_branch: int } (* next_branch= -1 iff not branchable *)
and connective =
    { mutable branches: constr list; 
      mutable next_branch: int; (* still needed? *) 
      mutable bloated: bool; (* true if result of pruning disjunction was different from pruning branches *)
      mutable changed: int list; (* branches that have been changed since last pruning, usually due to branching *)
      mutable wakeup_table: int list array;
      (* for each variable occurring in a branch, branches that contain this variable *)
      mutable maxconj_pos: bool; (* maximal conjunctive sub-constraint *)
      mutable maxconj_neg: bool
    }
and atomic =
    { obound: box; ibound: box; aconstr: constr } (* introduces new auxiliary variables with their bounds *)
and subset = 
    { subset_var1: variable; subset_var2: variable }
and redundant = 
    { original_constr: constr; redundant_constr: constr }
and quantbound =
    { var_trans: variable array; qbsubconstr: constr } (* a list old variables corresponding to new variables *)

val linear_relaxation: bool ref

val max_width: constr -> float 
(* returns the width of the widest interval occurring as a branchable quantification bound in the constraint,
   and zero if it does not contain a quantifier *)
(* val max_var: constr -> variable *) (* the variable of maximal index in #1 *) 
val neg_unary:predicateUnary -> predicateUnary
val neg: constr -> constr
(* the negatoin of the input *)

(*val complete_box: box->constr -> box*)
val copy: constr -> constr

val is_consistent_free: constr -> int (* free *) -> bool          (* constraint is consistent under the assumption that it has 'free' free variables *)
val is_consistent: constr -> box -> bool               (* consistency check for bounded constraints *)

val aux_consistent: constr -> bool               (* bounds for auxiliary variables consistent after full pruning? *)

type bconstr = constr * box                            (* still used? quantification instead *)

val max_width_bounded: bconstr -> float

val max_width_q: quantification -> float

val print_constraint:constr->int->int->unit      (* print #1 with indentation #2, quantified variables starting from #3 *)
val print_constr:constr->unit  
val print_bconstr:bconstr->unit

val contains_var: constr->variable-> bool        (* true iff #1 contains variable #2 (also if bound!) *)

val make_wakeup_table: constr list -> int -> int list array (* an array of #2 elements with k-th element the indices of all constraints in #1 that contain variable k *)

val build_connective: constr list -> int (* number of variables *) -> connective 

val connective_remove: connective -> int -> unit

(*
val consistent_marks: constr -> int -> bool -> int list -> bool 
                                                 (* true if #1 with #2 free variables is #3-consistent according to the marks in  #4, and within the constraint #1  *)
*)

val clear: constr -> unit  (* remove all information from earlier prunings, return to the state after parsing, replace ALL quantification bounds by [-inf, inf]; so: also quantification bounds from pruning are cleared! *)

val subst: constr -> int (* boolvar *) -> constr (* subst *) -> constr (* the result of  substituting 'subst' for all occurrences of 'boolvar' in ' constr' *)

(* val addvars: constr -> int (* free *) -> int (* before *) -> int (* after *) -> constr *)
  (* a copy of 'constr', 
     adding 'before' new free variables smaller than than the free variables,
     adding 'after' new free variables greater than than the free variables *)


val switch_branch_off: constr -> int (* num *) -> constr 
(* the result of switching off the branch 'num' of a connective, 
   i.e., replacing it by true in a conjunction and false in a disjunction,
   wakup_tables are not updated *)

val num_primitive: constr -> int

val add_branch: constr (* conjunction *) -> constr (* new_branch *) -> unit
  (* new_branch may NOT contain any new variables! *)

val decorate: constr -> unit 
  (* computes various information for constraint that are needed, but not immediately created by the parser:
     * relaxation tags: positive for root of maximal sub-constraint only containing conjunctions

     build wakeup tables also here instead of in parser? mean value constraints?
   *)
